/*
 * main_server.cpp
 *
 *  Created on: Nov 2, 2009
 *      Author: dportnoy
 */

#ifndef SERVER_HPP
#define SERVER_HPP

#include <utility>
#include <cstdlib>
#include <list>
#include <vector>
#include <map>

#include "csapp.h"
#include "constants.hpp"
#include "common.hpp"
#include "sync_list.hpp"
#include "gamestate.hpp"

using namespace std;

typedef struct player {
  int fd;
  struct sockaddr_in addr;
  char name[NAMELEN];
  struct game *game;

  //from game_server
  bool ingame;
  ship *s;
  int U;
  int D;
  int L;
  int R;
  int id;
  int weapon;
  int portal_hitpoints;
  float reverseTimer;
} player_t;

typedef struct game {
  int id;
  char name[NAMELEN];
  sync_list<player_t *> players;

  //from game_server
  bool started;
  
  float rotation_counter;
  int portal_direction;
  struct gamestate globalstate;
  player_t *players_arr[MAX_PLAYERS];
  player_t players_cache[MAX_PLAYERS];
  int numplayers;
  pthread_rwlock_t bllock;
  pthread_rwlock_t rmlock;
  pthread_rwlock_t lllock;
  pthread_rwlock_t itemlock;
  pthread_rwlock_t nlock;
  pthread_rwlock_t mlock;
  pthread_rwlock_t mslock;
  pthread_rwlock_t ulock;
  pthread_rwlock_t glock;
  pthread_rwlock_t elock;
  pthread_rwlock_t hlock;
  pthread_rwlock_t plock;
  pthread_rwlock_t playerl_lock;

  vector <bullet> bulletList;
  vector <removeItem> removeList;

  int num_bullets;
} game_t;

sync_list<game_t *> games;
sync_list<player_t *> players;
map<string, player_t *> player_map;

void kick_player(string msg, player_t *player);
int kick_player_from_game(player_t *player);
void send_disconnect_message(string msg, player_t *player);
void *thread(void *vargp);
void *game_thread(void *vargp);
int main(int argc, char **argv);

/* GAME SERVER FUNCTIONS */

/**@brief Plays the game
 *
 *
 * @param players An array of player_t types for all the players.
 *
 * @return Void.
 */
void playGame(player_t *players);

/**@brief Execution function for the listening threads.
 *
 *
 */
void *start_routine(void  *arg);

/**@brief Function to rotate the portal.
 *
 * Made modular so we could easily speed up rotation, 
 * switch directions, anything we want.
 *
 * @return Void.
 */
void rotate_portal(game_t *game);

/**@brief Returns the difference of two timevals.
 *
 *
 * @param x The later timeval.
 * @param y The earlier timeval.
 *
 * @return The difference in seconds
 */
float timeval_subtract (timeval *x, timeval *y);

/**@brief Sends a game state for this server.
 *
 * This will check the global game state and send back only things
 * that the players whose number is passed in needs to be concerned 
 * about. This means only objects on their screen, as well as a life
 * percentage for each player.
 *
 * @param player The number of the player whose screen we are getting.
 * @param players A pointer to the array fo player structures.
 *
 * @return Void. 
 */
void send_game_state(int, player_t **players, game_t *);

/**@brief Checks the keystroke states of a pleyers and updates
 *
 * @param num The number of the player to update.
 * @param player A pointer to this player's structure.
 *
 * @return void.
 */
void update_state(player_t *player, float time);

/**@brief Updates the states ob objects
 *
 * @param time The time elapsed since the last update.
 * 
 * @return Void.
 */
void update_object_states(float time, game_t *game);

/**@brief Updates a player's state.
 *
 * 
 *
 * @param player A pointer to the player structure to update.
 * @param field Which field we want to change.
 * @param action What to change the action to.
 *
 * @return Void.
 */
void update_player(player_t *, char *, char);

/**@brief Function to initialize a player.
 *
 * Creates a player's ship and initalizes all of his keystrokes
 * to down.
 *
 * @param num The player number.
 * @param player A pointer to the player struct.
 *
 * @return Returns 0 on success, and a -1 on failure.
 */
int player_initialize(int, player_t *, game_t *game);

/**@brief Initializes a bullet sets it up.
 *
 * Wrapper for bullet_create, and in case we have 
 * extra things to do other than create the initial
 * bullet.
 *
 * @param player A pointer to the player who owns this bullet.
 *
 * @return Void.
 */
void bullet_initialize(player_t * player);

/**@brief Functions to create a bullet object for a player 
 *
 * @param player A pointer to the player structure owning this  
 *
 * @return  A pointer to a new bullet.
 */
void bullet_create(player_t *player);

/**@brief Function to create a ship in a random position.
 *
 * @param player The player to create a ship for.
 */
ship* ship_create(int, int, game_t *game);

/**@brief Rotates a player's ship depending on the direction.
 *
 * @param s the ship
 * @param direction Positive value for clockwise, negative value for counter
 * @param time the time difference since last update
 *
 */
void ship_rotate(ship *s, int dir, float time);

/**@brief accelerates a ship (updates its velocity)
 *
 * @param s the ship
 * @param time the time difference since last update
 *
 */
void ship_accelerate(ship *s, float time);

void ship_deccelerate(ship *s, float time);

/**@brief Function to find the magnitude of a velocity.
 *
 * @param x X value of the vector.
 * @param y Y value fo the vector.
 * 
 * @return The magnitude of the vector.
 */
float magnitude(float x, float y);

/**@Will scale a vector to the desired length
 *
 * @param x A pointer to where the x value is stored.
 * @param y A pointer to where the y value is stored.
 * @param s A float that tells us how long to make theis vector.
 * @param cv The current velocity we are scaling from
 *
 * @return Void.
 */
void scale(float *x,float *y,float s,float cv);

/**@brief Moves a ship (updates its position)
 *
 * @param s the ship
 * @param time the time difference since last update
 *
 */
void move_ship(ship *s, float time);

/**@brief moves an object 
 *
 * @param position An array containing the x and y positions
 * @param x The change in x
 * @param y The change in y
 *
 */
void move_object(float *position,float x,float y);

/**@brief sets upper and lower to the upper and lower degrees of the portal
 *
 * @param portal The portal number
 * @param *upper A pointer to the float storing the upper value
 * @param *lower A pointer to the float storing the lower value
 * */
void getPortalAngles(int portal, float *upper, float *lower, game_t *game);

/**@brief Returns the angle from the top to center to this object
 *
 * @param x The x pos of the object
 * @param y The y pos of the object
 * */
float getAngle(float x, float y);

/**@brief Makes sure that the given position is on the screen. Wraps around to the other side if it is not
 *
 * @param position An array with the x and y positions
 * */
void wrap_around(float *position);

/**@brief Repositions a bullet after it goes from old_screen to new_screen
 *
 * @param old_screen The screen the object was on
 * @param new_screen The screen the object is now on
 * @param orientation The orientation of the object
 * @param position The position of the object
 * @param angle The angle from the top to center to this object
 * @param midline The angle of the middle of the portal it entered
 * */
void repositionObject(int old_screen, int new_screen, float *orientation, float *position, float angle, float midline, game_t *game);

float distance(float x1,float y1, float x2, float y2);

void initialize_special_weapon(int player, int screen, float *position, bool random, game_t *game);

void handle_projectile_collision(int player_id, int bullet_id, game_t *game);

void fire_special_weapon();

#endif
